open Types
exception Bug of string

module StringMap = Map.Make(String);;
module IntSet = Set.Make(struct type t = int32 let compare = compare end);;

let timing_only = false;;
let count_evals = ref 0;;
let verbose = ref 0;; (* 0 -- none, 1 -- debug, 2 -- verbose *)
let screen = ref true;;
let trunc = true;; (* truncate print arguments to lowest byte *)
let max_addr = 999;;
let max_space = Int64.of_int 200000;; (* otherwise too slow *)

type value_t =
  IVal of int32
  | FHndlT of string
  | FHndlB of string
  | Addr of string

type status_t =
  Bad of string
  | OK
  | Ret of int32
  | Brk
  | Cnt
  | Ext of int32

type resulte_t =
  GoodE of int32
  | ErrE of string

type resultp_t =
  GoodP of int32 * (int32 array)
  | ErrP of string

type delim_t =
  EOU of int32
  | EOS of int32
  | OSD of int32

type endpos_t =
  TxtDlm of IntSet.t * IntSet.t * IntSet.t
  | BinPos of int32 * bool

let char_of_int n =
  if Int32.compare n (Int32.of_int 9) == 0 ||
    Int32.compare n (Int32.of_int 10) == 0 ||
    (Int32.compare n (Int32.of_int 32) >= 0 && Int32.compare n (Int32.of_int 126) <= 0) then
    ("'" ^ (Char.escaped (Char.chr (Int32.to_int n))) ^ "'")
  else (Int32.to_string n)
;;

let screen_of_int n =
  if Int32.compare n (Int32.of_int 32) >= 0 && Int32.compare n (Int32.of_int 126) <= 0
  then Char.escaped (Char.chr (Int32.to_int n))
  else if Int32.compare n (Int32.of_int 9) == 0 then "\t"
  else if Int32.compare n (Int32.of_int 10) == 0 then "\n"
  else "\\" ^ (Int32.to_string n)
;;

let print_state (stack, global, heap, files, disk, buffer) =
  let print_varmap var = function
    IVal n -> Printf.eprintf "\tIVar %s = %s (0x%08lx)\n" var (char_of_int n) n
    | FHndlT hndl -> Printf.eprintf "\tFVar %s = FHndlT %s\n" var hndl
    | FHndlB hndl -> Printf.eprintf "\tFVar %s = FHndlB %s\n" var hndl
    | Addr addr -> Printf.eprintf "\tAVar %s = Addr %s\n" var addr in
  let print_heap addr data =
    let n = Int32.of_int (Array.length data) in
    let dl = Array.fold_left (fun nl n -> nl ^ " " ^ (char_of_int n)) "" data in
    let dl = if String.length dl > 40 then (String.sub dl 0 40) ^ "..." else dl in
    Printf.eprintf "\tAddr %s = (%ld) %s\n" addr n dl in
  let print_files hndl = (function
    (name, l, sou, TxtDlm (eou, eos, osd)) ->
      let fn = (fun n nl -> (char_of_int n) ^ " " ^ nl) in
      let set_eou = IntSet.fold fn eou "" in
      let set_eos = IntSet.fold fn eos "" in
      let set_osd = IntSet.fold fn osd "" in
      Printf.eprintf "\tFHndlT %s = (\"%s\", 0x%lx, 0x%lx, ({%s}, {%s}, {%s}))\n"
        hndl name l sou set_eou set_eos set_osd
    | (name, l, sou, BinPos (eou, little)) ->
      Printf.eprintf "\tFHndlB %s = (\"%s\", 0x%lx, 0x%lx, (0x%lx, %s))\n" hndl name l sou eou (if little then "Little" else "Big")) in
  let print_disk name data =
    let n = Int32.of_int (in_channel_length data) in
    Printf.eprintf "\t%s (0x%lx)\n" name n in
  if StringMap.is_empty stack then () else Printf.eprintf "Stack:\n"; StringMap.iter print_varmap stack;
  if StringMap.is_empty global then () else Printf.eprintf "Global:\n"; StringMap.iter print_varmap global;
  if StringMap.is_empty heap then () else Printf.eprintf "Heap:\n"; StringMap.iter print_heap heap;
  if StringMap.is_empty files then () else Printf.eprintf "Files:\n"; StringMap.iter print_files files;
  if StringMap.is_empty disk then () else Printf.eprintf "Disk:\n"; StringMap.iter print_disk disk;
  let contents = Array.fold_left (fun nl n -> nl ^ (screen_of_int n)) "" buffer in
  let len = Int32.of_int (String.length contents) in
  let contents =
    (if Int32.compare len (Int32.of_int 200) > 0 then
      let first50 = String.sub contents 0 100 in
      let last50 = String.sub contents (Int32.to_int (Int32.sub len (Int32.of_int 100))) 100 in
      first50 ^ "..." ^ last50
    else contents) in
  Printf.eprintf "Buffer: (%ld)\n%s\n" len contents
;;

let rec str_exp = function
  Int n -> char_of_int n
  | Var x -> x
  | BinOp (e1, op, e2) ->
      (let str_op = function
        Plus -> " + "
        | Minus -> " - "
        | Times -> " * "
        | Div -> " / "
        | Mod -> " % "
        | Eq -> " == "
        | Ne -> " != "
        | Lt -> " < "
        | Le -> " <= "
        | Gt -> " > "
        | Ge -> " >= "
        | And -> " && "
        | Or -> " || "
        | AndB -> " & "
        | OrB -> " | ";
        | ShiftR -> " >> "
        | ShiftL -> " << "
      in
      (str_exp e1) ^ (str_op op) ^ (str_exp e2))
  | Not e -> "!(" ^ (str_exp e) ^ ")"
  | NotB e -> "~(" ^ (str_exp e) ^ ")"
  | ARead (a, e) -> a ^ "[" ^ (str_exp e) ^ "]"
  | End (f, _) -> "end(" ^ f ^ ")"
  | Pos (f, _) -> "pos(" ^ f ^ ")"
  | Valid a -> "valid(" ^ a ^ ")"
;;

let str_dlm = function
  DArr a -> a
  | DInt n -> char_of_int n
;;

let rec str_stmt = function
  VWrite (x, e) -> x ^ " = " ^ (str_exp e) ^ ";"
  | Malloc (a, e, _) -> a ^ " = malloc(" ^ (str_exp e) ^ ");"
  | Free (a, _) -> "free(" ^ a ^ ");"
  | AWrite (a, e1, e2) -> a ^ "[" ^ (str_exp e1) ^ "] = " ^ (str_exp e2) ^ ";"
  | Call0 (x, q) -> x ^ " = " ^ q ^ "();"
  | Call (x, q, e) -> x ^ " = " ^ q ^ "(" ^ (str_exp e) ^ ");"
  | If (e, s1, s2) -> "if (" ^ (str_exp e) ^ ") {" ^ (str_stmts s1) ^ "} else {" ^ (str_stmts s2) ^ "}"
  | While (e, s) -> "while (" ^ (str_exp e) ^ ") {" ^ (str_stmts s) ^ "}"
  | Return e -> "return " ^ (str_exp e) ^ ";"
  | Break -> "break;"
  | Continue -> "continue;"
  | OpenT (f, name, _) -> f ^ " = open(" ^ name ^ ");" 
  | Seek (x, f, e, errcode) ->
      if errcode && x <> ""
      then x ^ " = seek(" ^ f ^ ", " ^ (str_exp e) ^ ");"
      else "seek(" ^ f ^ ", " ^ (str_exp e) ^ ");"
  | Read (x, f, _) -> x ^ " = read(" ^ f ^ ");"
  | Print e -> "print(" ^ (str_exp e) ^ ");"
  | Empty -> ";"

and str_stmts = function
  [] -> ""
  | s :: sl -> (str_stmt s) ^ (str_stmts sl)
;;

let arg = ref StringMap.empty;;
let body = ref StringMap.empty;;
let fr = ref StringMap.empty;;

let glodef = ref StringMap.empty;;
let heapdef = ref StringMap.empty;;
let filedef = ref StringMap.empty;;
let diskdef = ref StringMap.empty (* TODO *);;

let used_addr = ref 0;;
let used_hndl = ref 0;;
let used_space = ref Int64.zero;;

let allocate_stack size = true (* TODO overflow *);;

let update_var x u (stack, global, heap, files, disk, buffer) =
  if not (StringMap.mem x stack) && StringMap.mem x global then
    let global' = StringMap.add x u global in
    ((stack, global', heap, files, disk, buffer), OK)
  else
    let stack' = StringMap.add x u stack in
    ((stack', global, heap, files, disk, buffer), OK)
;;

let perform_var s x stack global fn =
  if StringMap.mem x stack then fn x stack
  else if StringMap.mem x global then fn x global
  else raise (Failure (x ^ " undefined (" ^ str_stmt s ^ ")"))
;;

let open_file f name text little errcode ((stack, global, heap, files, disk, buffer) as sigma) =
  if Sys.file_exists name then (* open-ok *)
    let in_channel = open_in name in
    let file_size = Int32.of_int (in_channel_length in_channel) in
    let hndl = "h" ^ (string_of_int !used_hndl) in
    used_hndl := !used_hndl + 1; (* TODO overflow *)
    let stack' =
      (if text then StringMap.add f (FHndlT hndl) stack
      else StringMap.add f (FHndlB hndl) stack) in
    let files' =
      (if text then StringMap.add hndl (name, Int32.zero, Int32.zero, TxtDlm (IntSet.empty, IntSet.empty, IntSet.empty)) files
      else StringMap.add hndl (name, Int32.zero, Int32.zero, BinPos (file_size, little)) files) in
    let disk' = StringMap.add name in_channel disk in
    ((stack', global, heap, files', disk', buffer), OK)
  else
    let msg = Printf.sprintf "File %s cannot be opened" name in
    if !verbose >= 1 then Printf.eprintf "%s\n" msg;
    (if errcode then
      let stack' = StringMap.add f (if text then (FHndlT "null") else (FHndlB "null")) stack in
      ((stack', global, heap, files, disk, buffer), OK)
    else (sigma, Bad msg) (* open-bad *))
;;

let rec parse_int data eou start width little =
  if Int32.compare width Int32.zero < 0 || Int32.compare width (Int32.of_int 4) > 0 then None
  else if Int32.compare (Int32.add start width) eou > 0 then None
  else if Int32.compare width Int32.zero == 0 then Some Int32.zero (* always nonnegative *)
  else
    (seek_in data (Int32.to_int start);
    let b = Int32.of_int (input_byte data) in
    match (parse_int data eou (Int32.succ start) (Int32.pred width) little, little) with
    (None, _) -> None
    | (Some i, true) ->
        let i' = Int32.add (Int32.shift_left i 8) b in
        if Int32.compare i' Int32.zero >= 0 then Some i' else None
    | (Some i, false) ->
        let i' = Int32.add (Int32.shift_left b (Int32.to_int (Int32.mul (Int32.pred width) (Int32.of_int 8)))) i in
        if Int32.compare i' Int32.zero >= 0 then Some i' else None)
;;

let read_byte data (eou, eos, osd) l =
  let n = Int32.of_int (in_channel_length data) in
  let delim = IntSet.union (IntSet.union eou eos) osd in
  if Int32.compare l n >= 0 then None
  else
    (seek_in data (Int32.to_int l);
    let b = Int32.of_int (input_byte data) in
    if IntSet.mem b delim then None
    else Some b)
;;

let rec to_delim data delim l =
  if Int32.compare l Int32.zero < 0 then to_delim data delim Int32.zero
  else match read_byte data delim l with
  None -> l
  | Some _ -> to_delim data delim (Int32.succ l)
;;

let next_delim hndl ((_, _, _, files, disk, _) as sigma) =
  match StringMap.find hndl files with
  (name, l, sou, TxtDlm (eou, eos, osd)) ->
    let data = StringMap.find name disk in
    let n = Int32.of_int (in_channel_length data) in
    if Int32.compare l n >= 0 then OSD n
    else 
      (let l' = to_delim data (eou, eos, osd) l in
      if Int32.compare l' n >= 0 then OSD n
      else
        (seek_in data (Int32.to_int l');
        let b = Int32.of_int (input_byte data) in
        if IntSet.mem b eou then EOU l'
        else if IntSet.mem b eos then EOS l'
        else OSD l'))
  | (_, _, _, BinPos _) ->
      if !verbose == 2 then print_state sigma;
      raise (Bug (hndl ^ " is not a text file hanlder"))
;;

let rec evale ((stack, global, heap, files, disk, _) as sigma) s = function
  Int n -> GoodE n (* int *)
  | Var x ->
      let find_var x varmap =
        (match StringMap.find x varmap with
        IVal n -> GoodE n (* ivar *)
        | _ -> raise (Failure (x ^ " is not an integer variable (" ^ str_stmt s ^ ")"))) in
      perform_var s x stack global find_var
  | BinOp (e1, op, e2) -> (* iop *)
      (match evale sigma s e1 with
      ErrE err -> ErrE err
      | GoodE u1 ->
          (match op with
          And -> if Int32.compare u1 Int32.zero == 0 then GoodE Int32.zero else
            (match evale sigma s e2 with
            ErrE err -> ErrE err
            | GoodE u2 -> if Int32.compare u2 Int32.zero <> 0 then GoodE Int32.one else GoodE Int32.zero)
          | Or -> if Int32.compare u1 Int32.zero <> 0 then GoodE Int32.one else
            (match evale sigma s e2 with
            ErrE err -> ErrE err
            | GoodE u2 -> if Int32.compare u2 Int32.zero <> 0 then GoodE Int32.one else GoodE Int32.zero)
          | _ ->
               (match evale sigma s e2 with
               ErrE err -> ErrE err
               | GoodE u2 ->
                   (match op with
                   Plus -> GoodE (Int32.add u1 u2)
                   | Minus -> GoodE (Int32.sub u1 u2)
                   | Times -> GoodE (Int32.mul u1 u2)
                   | Div -> if Int32.compare u2 Int32.zero <> 0 then GoodE (Int32.div u1 u2) else ErrE "Divide by zero"
                   | Mod -> if Int32.compare u2 Int32.zero <> 0 then GoodE (Int32.rem u1 u2) else ErrE "Modulo by zero"
                   | Eq -> if Int32.compare u1 u2 == 0 then GoodE Int32.one else GoodE Int32.zero
                   | Ne -> if Int32.compare u1 u2 <> 0 then GoodE Int32.one else GoodE Int32.zero
                   | Lt -> if Int32.compare u1 u2 < 0 then GoodE Int32.one else GoodE Int32.zero
                   | Gt -> if Int32.compare u1 u2 > 0 then GoodE Int32.one else GoodE Int32.zero
                   | Le -> if Int32.compare u1 u2 <= 0 then GoodE Int32.one else GoodE Int32.zero
                   | Ge -> if Int32.compare u1 u2 >= 0 then GoodE Int32.one else GoodE Int32.zero
                   | AndB -> GoodE (Int32.logand u1 u2)
                   | OrB -> GoodE (Int32.logor u1 u2)
                   | ShiftR -> GoodE (Int32.shift_right_logical u1 (Int32.to_int u2))
                   | ShiftL -> GoodE (Int32.shift_left u1 (Int32.to_int u2))
                   | _ -> raise (Bug "Impossible")))))
  | Not e -> (* iop *)
      (match evale sigma s e with
      ErrE err -> ErrE err
      | GoodE u -> if Int32.compare u Int32.zero <> 0 then GoodE Int32.zero else GoodE Int32.one)
  | NotB e -> (* iop *)
      (match evale sigma s e with
      ErrE err -> ErrE err
      | GoodE u -> GoodE (Int32.lognot u))
  | ARead (a, e) ->
      (match evale sigma s e with
      ErrE err -> ErrE err
      | GoodE u ->
          let find_arr a varmap =
            (match StringMap.find a varmap with
            Addr addr ->
              (if StringMap.mem addr heap then
                (let data = StringMap.find addr heap in
                let n = Int32.of_int (Array.length data) in
                if Int32.compare u Int32.zero < 0 || Int32.compare u n >= 0 then
                  ErrE (Printf.sprintf "Array read %s[%ld] out of bound" a u) (* ard-out *)
                else GoodE data.(Int32.to_int u) (* ard-ok *))
              else ErrE ("Read invalid adderss " ^ addr ^ " in " ^ a))
            | _ -> raise (Failure (a ^ " is not an array (" ^ str_stmt s ^ ")"))) in
          perform_var s a stack global find_arr)
  | End (f, errcode) ->
      let find_end f varmap =
        (match StringMap.find f varmap with
        FHndlT hndl ->
          (if StringMap.mem hndl files then
            (match StringMap.find hndl files with
            (name, l, _, TxtDlm (eou, eos, osd)) ->
              let data = StringMap.find name disk in
              (match read_byte data (eou, eos, osd) l with
              None -> GoodE (Int32.one) (* end-t *)
              | Some _ -> GoodE (Int32.zero) (* end-f *))
            | (_, _, _, BinPos _) ->
                if !verbose == 2 then print_state sigma;
                raise (Bug (hndl ^ " is not a text file handler")))
          else (if errcode then GoodE (Int32.minus_one) else ErrE "Test end of invalid handler"))
        | FHndlB hndl ->
            (if StringMap.mem hndl files then
              (match StringMap.find hndl files with
              (name, l, _, BinPos (eou, _)) ->
                if Int32.compare l eou == 0 then GoodE Int32.one (* end-t *)
                else GoodE Int32.zero (* end-f *)
              | (_, _, _, TxtDlm _) ->
                  if !verbose == 2 then print_state sigma;
                  raise (Bug (hndl ^ " is not a binary file handler")))
            else (if errcode then GoodE Int32.minus_one else ErrE "Test end of invalid handler"))
        | _ -> raise (Failure (f ^ " is not a file (" ^ str_stmt s ^ ")"))) in
      perform_var s f stack global find_end
  | Pos (f, errcode) ->  (* pos *)
      let find_pos f varmap =
        (match StringMap.find f varmap with
        FHndlT hndl ->
          (if StringMap.mem hndl files then
            let (_, l, _, _) = StringMap.find hndl files in
            GoodE l
          else (if errcode then GoodE Int32.minus_one else ErrE "Get position of invalid handler"))
        | FHndlB hndl ->
            (if StringMap.mem hndl files then
              let (_, l, _, _) = StringMap.find hndl files in
              GoodE l
            else (if errcode then GoodE Int32.minus_one else ErrE "Get position of invalid handler"))
        | _ -> raise (Failure (f ^ " is not a file"))) in
      perform_var s f stack global find_pos
  | Valid a ->
      let test_valid a varmap =
        (match StringMap.find a varmap with
        FHndlT hndl -> (if StringMap.mem hndl files then GoodE Int32.one else GoodE Int32.zero)
        | FHndlB hndl -> (if StringMap.mem hndl files then GoodE Int32.one else GoodE Int32.zero)
        | Addr addr -> (if StringMap.mem addr heap then GoodE Int32.one else GoodE Int32.zero)
        | _ -> raise (Failure (a ^ " is not an array or a file (" ^ str_stmt s ^ ")"))) in
      perform_var s a stack global test_valid
;;

let rec evals ((stack, global, heap, files, disk, buffer) as sigma) s =
  count_evals := !count_evals + 1;
  match s with
  VWrite (x, e) ->
    (match evale sigma s e with
    ErrE err -> (* vwr-bad *)
      let msg = Printf.sprintf "Bad expression in variable assignment (%s) (%s)" err (str_stmt s) in
      if !verbose >= 1 then Printf.eprintf "%s\n" msg;
      if !verbose == 2 then print_state sigma;
      (sigma, Bad msg)
    | GoodE u -> update_var x (IVal u) sigma (* vwr-ok *))
  | Malloc (a, e, errcode) ->
      (match evale sigma s e with
      ErrE err -> (* malloc-bad *)
        let msg = Printf.sprintf "Bad expression in malloc (%s) (%s)" err (str_stmt s) in
        if !verbose >= 1 then Printf.eprintf "%s\n" msg;
        (sigma, Bad msg)
      | GoodE u ->
          if Int32.compare u Int32.zero <= 0 then
            (let msg = Printf.sprintf "Bad size (%ld) in malloc %s (%s)" u a (str_stmt s) in
            if !verbose >= 1 then Printf.eprintf "%s\n" msg;
            if !verbose == 2 then print_state sigma;
            (if errcode then update_var a (Addr "null") sigma else (sigma, Bad msg) (* malloc-bad *)))
          else if !used_addr >= max_addr || Int64.compare (Int64.add !used_space (Int64.of_int32 u)) max_space >= 0 then
            (let msg = Printf.sprintf "Heap overflow requesting %ld, now using %Ld (%s)" u !used_space (str_stmt s) in
            if !verbose >= 1 then Printf.eprintf "%s\n" msg;
            if !verbose == 2 then print_state sigma;
            (if errcode then update_var a (Addr "null") sigma else (sigma, Bad msg) (* malloc-ovf *)))
          else (* allocate in global if exists in global *)
            (try
              let addr = Printf.sprintf "a%03d" !used_addr in
              let heap' = StringMap.add addr (Array.make (Int32.to_int u) Int32.zero) heap in
              used_addr := !used_addr + 1;
              used_space := Int64.add !used_space (Int64.of_int32 u);
              update_var a (Addr addr) (stack, global, heap', files, disk, buffer) (* malloc-ok *)
            with Out_of_memory | Invalid_argument _ ->
              let msg = Printf.sprintf "Heap overflow requesting %ld, now using %Ld (%s)" u !used_space (str_stmt s) in
              if !verbose >= 1 then Printf.eprintf "%s\n" msg;
              if !verbose == 2 then print_state sigma;
              (if errcode then update_var a (Addr "null") sigma else (sigma, Bad msg) (* malloc-ovf *))))
  | Free (a, errcode) -> (* x used only for receiving error code *)
      let free_arr a varmap =
        (match StringMap.find a varmap with
        Addr addr ->
          (if StringMap.mem addr heap then
            let n = Array.length (StringMap.find addr heap) in
            let heap' = StringMap.remove addr heap in
            used_space := Int64.sub !used_space (Int64.of_int n);
            let (sigma', _) = update_var a (Addr "null") (stack, global, heap', files, disk, buffer) in
            (sigma', OK)
          else
            let msg =
              (if !verbose >= 1 then Printf.sprintf "Free invalid address %s in %s (%s)" addr a (str_stmt s)
              else Printf.sprintf "Free invalid address in %s (%s)" a (str_stmt s)) in
            if !verbose >= 1 then Printf.eprintf "%s\n" msg;
            (sigma, Bad msg))
       | _ -> raise (Failure (a ^ " is not an array (" ^ str_stmt s ^ ")"))) in
      perform_var s a stack global free_arr
  | AWrite (a, e1, e2) ->
      (match evale sigma s e1 with
      ErrE err -> (* awr-bad *)
        let msg = Printf.sprintf "Bad expression in array index (%s) (%s)" err (str_stmt s) in
        if !verbose >= 1 then Printf.eprintf "%s\n" msg;
        (sigma, Bad msg)
      | GoodE u1 ->
          (match evale sigma s e2 with
          ErrE err -> (* awr-bad *)
            let msg = Printf.sprintf "Bad expression in array write (%s) (%s)" err (str_stmt s) in
            if !verbose >= 1 then Printf.eprintf "%s\n" msg;
            if !verbose == 2 then print_state sigma;
            (sigma, Bad msg)
          | GoodE u2 ->
              let write_arr a varmap =
                (match StringMap.find a varmap with
                Addr addr ->
                  (if StringMap.mem addr heap then
                    let data = StringMap.find addr heap in
                    let n = Int32.of_int (Array.length data) in
                    (if Int32.compare u1 Int32.zero < 0 || Int32.compare u1 n >= 0 then
                      (let msg = Printf.sprintf "Array write %s[%ld] out of bound (%s)" a u1 (str_stmt s) in
                      if !verbose >= 1 then Printf.eprintf "%s\n" msg;
                      if !verbose == 2 then print_state sigma;
                      (sigma, Bad msg) (* awr-out *))
                    else (* awr-ok *)
                      (let data' = Array.copy data in
                      Array.set data' (Int32.to_int u1) u2;
                      let heap' = StringMap.add addr data' heap in
                      ((stack, global, heap', files, disk, buffer), OK)))
                  else
                    (let msg =
                      (if !verbose >=1 then Printf.sprintf "Write invalid address %s in %s (%s)" addr a (str_stmt s)
                      else Printf.sprintf "Write invalid address in %s (%s)" a (str_stmt s)) in
                    if !verbose >= 1 then Printf.eprintf "%s\n" msg;
                    (sigma, Bad msg)))
                | _ -> raise (Failure (a ^ " is not an array (" ^ str_stmt s ^ ")"))) in
              perform_var s a stack global write_arr))
 | Call0 (x, q) ->
     (if not (StringMap.mem q !body && StringMap.mem q !fr) then raise (Failure (q ^ " undefined (" ^ str_stmt s ^ ")"))
     else if StringMap.mem q !arg then raise (Failure ("Function " ^ q ^ " argument missing (" ^ str_stmt s ^ ")"))
     else
       (if allocate_stack (StringMap.find q !fr) then
         (match evalsl (StringMap.empty, global, heap, files, disk, buffer) (StringMap.find q !body) with
         ((_, _, _, files', _, _), Bad err) -> ((stack, global, heap, files', disk, buffer), Bad err) (* fn-body *)
         | ((_, global', heap', files', disk', buffer'), Ret retval) ->
             update_var x (IVal retval) (stack, global', heap', files', disk', buffer') (* fn-prgr *)
         | (_, Ext _) as phi -> phi
         | (_, Brk) -> raise (Failure ("Redundant break statement in " ^ q))
         | (_, Cnt) -> raise (Failure ("Redundant continue statement in " ^ q))
         | _ -> raise (Failure ("Missing return statement in " ^ q)))
       else
         let msg = "Stack overflow" in
         if !verbose >= 1 then Printf.eprintf "%s\n" msg;
         (sigma, Bad msg) (* fn-ovf *)))
 | Call (x, q, e) -> (* not an expression because of side effects on files XXX really? *)
     (match evale sigma s e with
     ErrE err -> (* fn-arg *)
       let msg = Printf.sprintf "Bad expression in function argument (%s) (%s)" err (str_stmt s) in
       if !verbose >= 1 then Printf.eprintf "%s\n" msg;
       (sigma, Bad msg)
     | GoodE u ->
         (if not (StringMap.mem q !body && StringMap.mem q !fr) then raise (Failure (q ^ " undefined (" ^ str_stmt s ^ ")"))
         else if not (StringMap.mem q !arg) then raise (Failure ("Function " ^ q ^ " argument redundant (" ^ str_stmt s ^ ")"))
         else
           (if allocate_stack (StringMap.find q !fr) then
             let inside = StringMap.add (StringMap.find q !arg) (IVal u) StringMap.empty in
             (match evalsl (inside, global, heap, files, disk, buffer) (StringMap.find q !body) with
             ((_, _, _, files', _, _), Bad err) -> ((stack, global, heap, files', disk, buffer), Bad err) (* fn-body *)
             | ((_, global', heap', files', disk', buffer'), Ret retval) ->
                 update_var x (IVal retval) (stack, global', heap', files', disk', buffer') (* fn-prgr *)
             | (_, Ext _) as phi -> phi
             | (_, Brk) -> raise (Failure ("Redundant break statement in " ^ q))
             | (_, Cnt) -> raise (Failure ("Redundant continue statement in " ^ q))
             | _ -> raise (Failure ("Missing return statement in " ^ q)))
           else
             let msg = Printf.sprintf "Stack overflow (%s)" (str_stmt s) in
             if !verbose >= 1 then Printf.eprintf "%s\n" msg;
             (sigma, Bad msg) (* fn-ovf *))))
 | If (e, sl1, sl2) ->
     (match evale sigma s e with
     ErrE err ->
       let msg = Printf.sprintf "Bad expression in condition (%s) (%s)" err (str_stmt s) in
       if !verbose >= 1 then Printf.eprintf "%s\n" msg;
       if !verbose == 2 then print_state sigma;
       (sigma, Bad msg) (* if-bad *)
     | GoodE u ->
         if Int32.compare u Int32.zero == 0 then evalsl sigma sl2 (* if-f *)
         else evalsl sigma sl1 (* if-t *))
 | While (e, sl) ->
     (match evale sigma s e with
     ErrE err ->
       let msg = Printf.sprintf "Bad expression in loop predicate (%s) (%s)" err (str_stmt s) in
       if !verbose >= 1 then Printf.eprintf "%s\n" msg;
       (sigma, Bad msg) (* while-bad *)
     | GoodE u ->
         if Int32.compare u Int32.zero == 0 then (sigma, OK) (* while-end *)
         else 
           (match evalsl sigma sl with
           (sigma', Bad err) -> (sigma', Bad err) (* while-body *)
           | (sigma', OK) -> evals sigma' (While (e, sl)) (* while-prgr *)
           | (sigma', Ret retval) -> (sigma', Ret retval)
           | (sigma', Brk) -> (sigma', OK) (* while-brk *)
           | (sigma', Cnt) -> evals sigma' (While (e, sl)) (* while-cnt *)
           | (_, Ext _) as phi -> phi))
 | Return e ->
     (match evale sigma s e with
     ErrE err ->
       let msg = Printf.sprintf "Bad expression in return (%s) (%s)" err (str_stmt s) in
       if !verbose >= 1 then Printf.eprintf "%s\n" msg;
       (sigma, Bad msg)
     | GoodE u -> (sigma, Ret u))
 | Break -> (sigma, Brk)
 | Continue -> (sigma, Cnt)
 | OpenT (f, name, errcode) ->
     (if StringMap.mem f stack then (* only check stack. always open in stack, no matter if exists in global *)
       raise (Failure (f ^ " already in use (" ^ str_stmt s ^ ")"))
     else open_file f name true false errcode sigma)
 | Seek (x, f, e, errcode) -> (* x used only for receiving error code *)
     (match evale sigma s e with
     ErrE err ->
       let msg = Printf.sprintf "Bad expression in seek (%s) (%s)" err (str_stmt s) in
       if !verbose >= 1 then Printf.eprintf "%s\n" msg;
       (sigma, Bad msg) (* sk-bad *)
     | GoodE u ->
         let seek_file f varmap =
           (match StringMap.find f varmap with
           FHndlT hndl ->
             (if StringMap.mem hndl files then
               (match StringMap.find hndl files with
               (name, l, sou, TxtDlm (eou, eos, osd)) ->
                 let data = StringMap.find name disk in
                 if Int32.compare u sou < 0 then (None, Some "Seek before the start of current input unit") (* sk-l *)
                 else if Int32.compare u l <= 0 || Int32.compare u (to_delim data (eou, eos, osd) l) <= 0 then
                   (Some (StringMap.add hndl (name, u, sou, TxtDlm (eou, eos, osd)) files), None) (* sk-ok *)
                 else (None, Some "Seek past the end of current input unit or file") (* sk-r *)
               | (_, _, _, BinPos _) ->
                   if !verbose == 2 then print_state sigma;
                   raise (Bug (hndl ^ " is not a text file handler")))
             else (None, Some "Seek invalid file handler"))
           | FHndlB hndl ->
               (if StringMap.mem hndl files then
                 (match StringMap.find hndl files with
                 (name, l, sou, BinPos (eou, little)) ->
                   if Int32.compare u sou < 0 then (None, Some "Seek before the start of current input unit") (* sk-l *)
                   else if Int32.compare u l <= 0 || Int32.compare u eou <= 0 then
                     (Some (StringMap.add hndl (name, u, sou, BinPos (eou, little)) files), None) (* sk-ok *)
                   else (None, Some "Seek past the end of current input unit or file") (* sk-r *)
                 | (_, _, _, TxtDlm _) ->
                     if !verbose == 2 then print_state sigma;
                     raise (Bug (hndl ^ " is not a binary file handler")))
               else (None, Some "Seek invalid file handler"))
           | _ -> raise (Failure (f ^ " is not a file (" ^ str_stmt s ^ ")"))) in
         if !verbose == 2 then Printf.eprintf "seek(%s, %ld)\n" f u;
         (match perform_var s f stack global seek_file with
         (_, Some err) -> (if errcode then update_var x (IVal Int32.minus_one) sigma else (sigma, Bad err))
         | (Some files', _) ->
             (if errcode then update_var x (IVal u) (stack, global, heap, files', disk, buffer)
             else ((stack, global, heap, files', disk, buffer), OK))
         | _ -> raise (Bug "Impossible")))
 | Read (x, f, errcode) -> (* Not an expression because of side effects*)
     let read_file f varmap =
       (match StringMap.find f varmap with
       FHndlT hndl ->
         (if StringMap.mem hndl files then
           (match StringMap.find hndl files with
           (name, l, sou, TxtDlm (eou, eos, osd)) ->
             let data = StringMap.find name disk in
             (match read_byte data (eou, eos, osd) l with
             None -> None
             | Some b ->
                 let files' = StringMap.add hndl (name, Int32.succ l , sou, TxtDlm (eou, eos, osd)) files in
                 Some (b, files'))
           | (_, _, _, BinPos _) ->
               if !verbose == 2 then print_state sigma;
               raise (Bug (hndl ^ " is not a text file handler")))
         else None)
       | FHndlB hndl ->
           (if StringMap.mem hndl files then
             (match StringMap.find hndl files with
             (name, l, sou, BinPos (eou, little)) ->
               let data = StringMap.find name disk in
               if Int32.compare l eou >= 0 then None
               else
                 (seek_in data (Int32.to_int l);
                 let b = Int32.of_int (input_byte data) in
                 let files' = StringMap.add hndl (name, Int32.succ l, sou, BinPos (eou, little)) files in
                 Some (b, files'))
             | (_, _, _, TxtDlm _) ->
                 if !verbose == 2 then print_state sigma;
                 raise (Bug (hndl ^ " is not a binary file handler")))
           else None)
       | _ -> raise (Failure (f ^ " is not a file (" ^ str_stmt s ^ ")"))) in
     (match perform_var s f stack global read_file with
     None ->
       let msg = Printf.sprintf "Read invalid handler or past the end of current input unit (%s)" (str_stmt s) in
       if !verbose >= 1 then Printf.eprintf "%s\n" msg;
       if !verbose == 2 then print_state sigma;
       (if errcode then update_var x (IVal Int32.minus_one) sigma else (sigma, Bad msg) (* frd-out *))
     | Some (b, files') -> update_var x (IVal b) (stack, global, heap, files', disk, buffer) (* frd-ok *))
 | Print e ->
     let prt_int n =
       (let buffer' = Array.append buffer (Array.make 1 n) in
       ((stack, global, heap, files, disk, buffer'), OK)) in
     let rec prt_arr a i =
       (if i >= Array.length a then a
       else if Int32.compare a.(i) Int32.zero == 0 then (
          if i == 0 then Array.make 0 Int32.zero
          else Array.sub a 0 i)
       else prt_arr a (i+1)) in
     let prt_arr_var =
       (match e with
       Var x ->
         let is_arr_var x varmap =
           (match StringMap.find x varmap with
           Addr n -> (true, x)
           | _ -> (false, "")) in
         perform_var s x stack global is_arr_var
       | _ -> (false, "")) in
     (match prt_arr_var with
     (true, x) ->
       (match perform_var s x stack global StringMap.find with
       Addr addr ->
         (if StringMap.mem addr heap then
           let data = StringMap.find addr heap in
           let buffer' = Array.append buffer (if !screen then (prt_arr data 0) else data) in
           (* if print screen, then print single '\0's but skip trailing bytes in arrays *)
           (* if print file, then print every byte given *)
           ((stack, global, heap, files, disk, buffer'), OK)
         else
           let msg = Printf.sprintf "Invalid address %s in %s (%s)" addr x (str_stmt s) in
           if !verbose >= 1 then Printf.eprintf "%s\n" msg;
           (sigma, Bad msg))
       | _ -> raise (Bug ("expecting an array variable")))
     | (false, _) ->
         (match evale sigma s e with
         ErrE err ->
           let msg = Printf.sprintf "Bad expression in print (%s) (%s)" err (str_stmt s) in
           if !verbose >= 1 then Printf.eprintf "%s\n" msg;
           (sigma, Bad msg)
         | GoodE u -> prt_int u))
 | Empty -> (sigma, OK)

and evalsl sigma = function
  [] -> (sigma, OK)
  | s::sl ->
      (match evals sigma s with
      (sigma', Bad err) ->
        if !verbose == 2 then Printf.eprintf "[Runtime error]: %s\n" err;
        (sigma', Bad err) (* seq-bad *)
      | (sigma', OK) -> evalsl sigma' sl (* seq-prgr *)
      | (_, Ret _) as phi -> phi
      | (_, Brk) as phi -> phi
      | (_, Cnt) as phi -> phi
      | (_, Ext _) as phi -> phi)
;;

let rec evalp = function
  Main (body) ->
    (match evalsl (StringMap.empty, !glodef, !heapdef, !filedef, !diskdef, Array.make 0 Int32.zero) body with
    (_, Bad err) -> ErrP err
    | ((_, _, _, _, _, buffer') as sigma', Ret retval) ->
        if !verbose >= 1 then print_state sigma';
        GoodP (retval, buffer')
    | (_, Brk) -> raise (Failure ("Redundant break statement in main"))
    | (_, Cnt) -> raise (Failure ("Redundant continue statement in main"))
    | (sigma', Ext errno) ->
        if !verbose >= 1 then print_state sigma';
        ErrP (Printf.sprintf "Exit with errno %ld" errno)
    | (_, OK) -> raise (Failure ("Missing return statement in main")))
  | Fn0Def (q, qbody, prog) ->
      body := StringMap.add q qbody !body;
      fr := StringMap.add q 100 !fr; (* TODO *)
      if !verbose == 2 then Printf.eprintf "func %s ()\n" q;
      evalp prog
  | FnDef (q, qarg, qbody, prog) ->
      arg := StringMap.add q qarg !arg;
      body := StringMap.add q qbody !body;
      fr := StringMap.add q 100 !fr; (* TODO *)
      if !verbose == 2 then Printf.eprintf "func %s (%s)\n" q qarg;
      evalp prog
  | GloDef (glo, prog) ->
      (match glo with
      FDefT (f, name, errcode) ->
          (if Sys.file_exists name then (* open-ok *)
            let in_channel = open_in name in
            let hndl = "h" ^ (string_of_int !used_hndl) in
            used_hndl := !used_hndl + 1; (* TODO overflow *)
            glodef := StringMap.add f (FHndlT hndl) !glodef;
            filedef := StringMap.add hndl (name, Int32.zero, Int32.zero, TxtDlm (IntSet.empty, IntSet.empty, IntSet.empty)) !filedef;
            diskdef := StringMap.add name in_channel !diskdef;
            evalp prog
          else
            let msg = Printf.sprintf "File %s cannot be opened" name in
            if !verbose >= 1 then Printf.eprintf "%s\n" msg;
            (if errcode then
              (glodef := StringMap.add f (FHndlT "null") !glodef;
              evalp prog)
            else ErrP msg (* open-bad *))))
;;

let _ =
  if Array.length Sys.argv <> 3 then
    (Printf.eprintf "Usage:\n\t%s [source_file] [verbose level 0-2]\n" Sys.argv.(0); exit 1)
  else if (int_of_string Sys.argv.(2) < 0 ||
           int_of_string Sys.argv.(2) > 2) then
    (Printf.eprintf "Usage:\n\t%s [source_file] [verbose level 0-2]\n" Sys.argv.(0); exit 1)
  else
    screen := true;
    verbose := int_of_string Sys.argv.(2);
    let in_channel = open_in Sys.argv.(1) in
    let lexbuf = Lexing.from_channel in_channel in
    try
      while true do
        let program = Parser.prog Lexer.token lexbuf in
        let time_start = Unix.gettimeofday() in
        let result = evalp program in
        let time_end = Unix.gettimeofday() in
        (match result with
        ErrP err -> Printf.eprintf "[Aborted with error]: %s.\n" err;
        | GoodP (u, buffer) ->
            let duration = 1000. *. (time_end -. time_start) in
            if timing_only then (Printf.printf "%f ms\t%d stmt" duration !count_evals)
            else 
              (let to_byte n = if trunc then Int32.logand n (Int32.of_int 255) else n in
              if !screen then (* print to screen *)
                let contents = Array.fold_left (fun nl n -> nl ^ (screen_of_int (to_byte n))) "" buffer in
                Printf.printf "%s\n" contents;
              else (* write to file *)
                let output = open_out "output.bin" in
                Array.iter (fun n -> output_char output (Char.chr (Int32.to_int (to_byte n)))) buffer;
                close_out output;);
              Printf.eprintf "[Program return value]: %ld.\n" u;
              );
        flush stderr; flush stdout
      done
    with
      Lexer.Eof ->
        if !verbose >= 1 then Printf.eprintf "[End of code]\n";
        close_in in_channel;
        exit 0
    | Lexer.Unexpected ->
        let line = !Lexer.lineno in
        let tok = Lexing.lexeme lexbuf in
        Printf.eprintf "[Syntax error]: Unexpected character \"%s\" (line %d).\n" tok line;
        exit 1
    | Parsing.Parse_error ->
        let line = !Lexer.lineno in
        let tok = Lexing.lexeme lexbuf in
        Printf.eprintf "[Syntax error]: Expecting something else at \"%s\" (line %d).\n" tok line;
        exit 1
    | Failure err ->
        Printf.eprintf "[Syntax error]: %s.\n" err;
        exit 1
    | Not_found ->
        Printf.eprintf "[Syntax error]: Identifier not found.\n";
        exit 1
    | Bug err ->
        Printf.eprintf "[Implementation error]: Your code triggered an unexpected bug in the interpreter (%s).\n" err;
        exit 1
;;
